package de.uks.plm.ftglb.actor.dataStructures;

/**
 * This class describes a timeout-entry. Every entry constists of a place
 * (the place we are waiting for), the reason for the timout, an check-interval
 * and the time of the earliest next check.
 */

import x10.compiler.Inline;

public class Timeout implements Comparable[Timeout]
{
    public static val CHECK_INTERVAL:Long = 100;

    /** Reasons: */
    public static val REG_ACK:Long = 1;
    public static val STL_ACK:Long = 2;
    public static val TO_ACK:Long = 4;
    public static val IA_ACK:Long = 8;
    public static val RST_ACK:Long = 16;
    public static val LNK_RES:Long = 32;
    public static val GOT_ACK:Long = 64;
    public static val GIVE:Long = 128;
    public static val BF_ACK:Long = 256;
    public static val BV_END:Long = 512;
    public static val F_END:Long = 1024;
    public static val PROCESS_JOINS:Long = 2048;

    /** Place to check */
    public val placeId:Long;

    /** Reason to check */
    public val reason:Long;

    /** (Earliest) Time of the next check */
    private var nextCheck:Long;

    /** Check-intervals */
    public val interval:Long;

    public def this(val placeId:Long, val reason:Long)
        : Timeout
    {
        this(placeId
             , reason
             , System.currentTimeMillis() + Timeout.CHECK_INTERVAL
             , Timeout.CHECK_INTERVAL);
    }

    public def this(val placeId:Long, val reason:Long, val interval:Long)
        : Timeout
    {
        this(placeId
             , reason
             , System.currentTimeMillis() + Timeout.CHECK_INTERVAL
             , interval);
    }

    public def this(val placeId:Long
                    , val reason:Long
                    , val nextCheck:Long
                    , val interval:Long) : Timeout
    {
        this.placeId = placeId;
        this.reason = reason;
        this.nextCheck = nextCheck;
        this.interval = interval;
    }

    /**
     * Checks the specified place and sets the next check.
     * Role: any
     * Sink: yes
     * @return true, if the checked place is still alive
     *         false, otherwise
     */
    @Inline public def check() : Boolean
    {
        val place:Place = Place(this.placeId);
        if (place.isDead())
        {
            return (false);
        }
        else // if (!place.isDead())
        {
            this.setNextCheck(System.currentTimeMillis() + this.interval);
            return (true);
        }
    }
    /**
     * Define comparator, so a list of comparator is always sorted (smallest
     * nextCheck-value first, if equal, smaller place-number first).
     */
    @Inline public def compareTo(val that:Timeout):Int
    {   // Compare next check time
        if (this.nextCheck < that.nextCheck)
        {
            return -1n;
        }
        if (this.nextCheck > that.nextCheck)
        {
            return 1n;
        }
        // Compare place-id
        if (this.placeId < that.placeId)
        {
            return -1n;
        }
        if (this.placeId < that.placeId)
        {
            return 1n;
        }
        // Equal
        return 0n;
    }

    /**
     * Get the next check time (as long)
     */
    @Inline public def getNextCheck() : Long
    {
        return (this.nextCheck);
    }

    /**
     * Set the next check time (as long)
     */
    @Inline public def setNextCheck(val nextCheck:Long) : void
    {
        this.nextCheck = nextCheck;
    }

    @Inline public def equals(val that:Any) : Boolean
    {
        if (that instanceof Timeout)
        {
            return (this.equals(that as Timeout));
        }
        return (false);
    }

    @Inline public def equals(val that:Timeout) : Boolean
    {
        return ((this.placeId == that.placeId) && (this.reason == that.reason));
    }

    @Inline public def toString()
    {
        return (  "Timeout-> place = " + Place(this.placeId)
                + ", reason = " + this.reason);
    }
}
