package de.uks.plm.ftglb.actor.debugClass;

import x10.compiler.NonEscaping;

import x10.util.ArrayList;
import x10.util.List;
import x10.util.Option;
import x10.util.OptionsParser;

import de.uks.plm.ftglb.actor.GLB;
import de.uks.plm.ftglb.actor.GLBParameters;
import de.uks.plm.ftglb.actor.TaskBag;
import de.uks.plm.ftglb.actor.TaskQueue;

import de.uks.plm.utils.AssertPrinter;
import de.uks.plm.utils.RingBuffer;

public class Debug implements TaskQueue[Debug, Long]
{
    private val buffer:RingBuffer[Long];
    private var result:Long = 0;

    public static def main(args:Rail[String]) : void
    {
        val opts = new OptionsParser
            (args
             , new Rail[Option]()
             , [
                Option("s", "", "Initial tasks")
                , Option("n"
                         , ""
                         ,  "Number of nodes to process before probing. "
                          + "Default 200.")
                , Option("w", "", "Number of thieves to send  out. Default 1.")
                , Option("l", "", "Base of the lifeline")
                , Option("m", "", "Max potential victims")
                , Option("k", "", "backup-cycle")
                , Option("r", "", "monitor-cycle")
                , Option("v", "", "Verbose. Default 0 (no).")]);

        val s = opts("-s", 100);
        val n = opts("-n", 511n);
        val l = opts("-l", 32n);
        val k = opts("-k", 10n);
        val m = opts("-m", 1024n);
        val r = opts("-r", 128n);
        val verbose = opts("-v", GLBParameters.SHOW_RESULT_FLAG);
        val P = Place.numPlaces();

        var z0:Int = 1n;
        var zz:Int = l;
        while (zz < P)
        {
            z0++;
            zz *= l;
        }
        val z = z0;

        val w = opts("-w", z);

        Console.OUT.println("places=" + P +
                            "   s=" + s +
                            "   w=" + w +
                            "   n=" + n +
                            "   l=" + l + 
                            "   m=" + m + 
                            "   z=" + z);

        val init:() => Debug = () =>
                               {
                                   return (new Debug(s));
                               };

        val params:GLBParameters = new GLBParameters(n, w, l, z, m, k, r, verbose);
        val glb = new GLB[Debug,Long](init , params , true);

        val start:() => void = () =>
                               {
                                   (glb.taskQueue()).init(s);
                               };

        val result:Rail[Long] = glb.run(start);
    }

    public def this(val bound:Long) : Debug
    {
        this.buffer = new RingBuffer[Long]();
        this.result = 0;
    }

    public def this(val data:Rail[Long], val result:Long) : Debug
    {
        this.buffer = new RingBuffer[Long](data);
        this.result = result;
    }

    public def this(val buffer:RingBuffer[Long], val result:Long) : Debug
    {
        this.buffer = buffer;
        this.result = result;
    }

    public def this(val that:Debug) : Debug
    {
        this.buffer = new RingBuffer[Long](that.buffer);
        this.result = that.result;
    }

    @NonEscaping public final def init(val bound:Long)
    {
        this.buffer.pushTop(bound);
    }

    public def process(val n:Long) : Boolean
    {
        for (var i:Long = 0; (i < n) && this.buffer.size() > 0; ++i)
        {
            val value:Long = (this.buffer.popTop() - 1);
            ++this.result;
            if (0 <= value)
            {
                for (j:Long in 0..(value))
                {
                    this.buffer.pushTop(j);
                }
            }
        }
        return (this.buffer.size() > 0);
    }

    public def split() : Bag
    {
        val otherSize:Long = this.buffer.size() / 2;
        val mySize:Long = this.buffer.size() - otherSize;
        if (otherSize > 0)
        {
            
            val data:Rail[Long] = this.buffer.getFromBottom(otherSize);
            val loot:Bag = new Bag(data);
            AssertPrinter.println(here + ": split = " + loot);
            AssertPrinter.println(here + ": my buffer = " + this.buffer);
            return (loot);
        }
        return(null as Bag);
    }

    public def mergeAtBottom(val loot:Bag) : void
    {
        AssertPrinter.println(here + ": loot = " + loot);
        this.buffer.pushBottom(loot.rail);
        AssertPrinter.println(here + ": buffer = " + this.buffer);
    }

    public def mergeAtBottom(val loot:TaskBag) : void
    {
        this.merge(loot as Bag);
    }

    public def merge(val loot:Bag) : void
    {
        this.mergeAtBottom(loot);
    }

    public def merge(val loot:TaskBag) : void
    {
        this.mergeAtBottom(loot as Bag);
    }

    public def mergeAtTop(val loot:Bag) : void
    {
        AssertPrinter.println(here + ": split = " + loot);
        this.buffer.pushTop(loot.rail);
        AssertPrinter.println(here + ": buffer = " + this.buffer);
    }

    public def mergeAtTop(val loot:TaskBag) : void
    {
        this.mergeAtTop(loot as Bag);
    }

    public def count() : Long
    {
        return (this.result);
    }

    public def getResult() : Result
    {
        return (new Result(this.result));
    }

    public def printLog() : void {}

    public def size() : Long
    {
        return (this.buffer.size());
    }

    public def clone() : Debug
    {
         return (new Debug(this));
    }

    public def getQueueFromTop(val count:Long) : Debug
    {
        return (new Debug(this.buffer.peekFromTop(count)
                          , this.result));
    }

    public def getQueueFromBottom(val count:Long) : Debug
    {
        return (new Debug(this.buffer.peekFromBottom(count)
                          , this.result));
    }

    public def getTopElement() : Debug
    {
        val data:Rail[Long] = new Rail[Long](1);
        data(0) = this.buffer.peekTop();
        return (new Debug(data, this.result));
    }

    public def removeFromTop(val count:Long) : void
    {
        this.buffer.getFromTop(count);
    }

    public def removeFromBottom(val count:Long) : void
    {
        this.buffer.getFromBottom(count);
    }

    public def mergeAtTop(val that:Debug) : void
    {
        this.buffer.mergeAtTop(that.buffer);
    }

    public def mergeAtBottom(val that:Debug) : void
    {
        this.buffer.mergeAtBottom(that.buffer);
    }

    public def setResult(val that:Debug) : void
    {
        this.result = that.result;
    }

    public def mergeResult(val that:Debug) : void
    {
        this.result += that.result;
    }

    public def mergeResult(val that:TaskQueue[Debug, Long])
    {
        this.mergeResult(that as Debug);
    }

    public def toString() : String
    {
        return (this.buffer.toString());
    }
}
