package de.uks.plm.ftglb.actor.managers;

import x10.compiler.Inline;
import x10.compiler.NonEscaping;
import x10.compiler.Uncounted;

import x10.util.ArrayList;
import x10.util.List;
import x10.util.Map;
import x10.util.Pair;
import x10.util.Set;

import de.uks.plm.ftglb.actor.TaskQueue;
import de.uks.plm.ftglb.actor.Worker;

import de.uks.plm.ftglb.actor.dataStructures.BackupData;

import de.uks.plm.utils.AssertPrinter;

import de.uks.plm.ftglb.actor.managers.LogManager;
import de.uks.plm.ftglb.actor.logEvents.*;

public final class BackupReceiverManager[Queue, R] {Queue <: TaskQueue[Queue, R]}
{
    /**
     * The worker, to which this reciever belongs, i.e. the worker, who receives
     * the backups.
     */
    val worker:Worker[Queue, R];

    /**
     * The backup-queue
     */
    public var backup:Queue;

    /**
     * The backup-log
     */
    public var backupLog:Long;

    /**
     * A list of thieves. used, while in a touchy situation
     */
    public var thieves:List[Long];

    public var tan:Long;
    
    public def this(val worker:Worker[Queue, R])
        : BackupReceiverManager[Queue, R]
    {
        this.worker = worker;
        this.backup = (null as Queue);
        this.backupLog = this.worker.me;
        this.thieves = new ArrayList[Long]();
        this.tan = Long.MAX_VALUE;
    }

    /**
     * Receive a backup and update the current backup according to kind and
     * cause.
     * Role: back
     * sink: no, we have to send an ack to the source-place
     */
    @Inline public def update(val data:BackupData[Queue, R]) : void
    {
        var leReceiveBackup:LEreceiveBackup = new LEreceiveBackup(System.nanoTime());
        worker.logManager.startLE(leReceiveBackup);
        leReceiveBackup.kindBack = data.kind;
        leReceiveBackup.causeBack = data.cause;

        AssertPrinter.println(  here + ": current backup has size: "
                              + (null == this.backup
                                 ? 0
                                 : this.backup.size())
                              + "\nnewQueue: " + (null == data.backup
                                                  ? "null"
                                                  : data.backup.size())
                              + "\nBackup-Type = " + data.kind);
        leReceiveBackup.sourceId = data.sourceId;
        if (this.worker.forth() == data.sourceId)
        {
            // if (BackupData.KIND_COMP == data.kind)
            // {
                leReceiveBackup.openLinks = this.fullUpdate(data);
            // }
            // else // if (BackupData.KIND_INCR == data.kind)
            // {
            //     this.incrementalUpdate(data);
            //     // AssertPrinter.println("[ Worker " + here + "]: PANIC!");
            // }

            AssertPrinter.println(  "[Receiver " + here
                                  + "]: newBackup-Size(is-state) = "
                                  + this.backup.size()
                                  + ", new result = "
                                  + (this.backup !=  null ? this.backup.getResult().getResult() : "null"));

            this.respond(data);
        }
        else
        {   // TODO: special situations
            AssertPrinter.println(  "[Receiver " + here + "]: Uh-oh. "
                                  + data.sourceId + "sent, but I expected "
                                  + this.worker.forth() + " to send.");
        }
        worker.logManager.endLE();
    }

    /**
     * Interpret the data as full update.
     * Role: backup
     * sink: yes
     */
    // Return value is for logging purposes.
    @Inline public def fullUpdate(val data:BackupData[Queue, R]) : Long
    {
        this.backup = data.backup;

        if (-1 < data.log)
        {
            this.backupLog = data.log;
        }

        if (Long.MAX_VALUE > data.tan)
        {
            this.tan = data.tan;
        }
        else
        {
            this.tan = Long.MAX_VALUE;
        }

        if (BackupData.CAUSE_IA == data.cause)
        {
            assert (this.worker.linkManager.isEmpty())
                    : "Writing IA-Backup, but linkManager is not emoty.";
        }
        var countOpenLinks:Long = 0; //For logging.
        val linkMap:Map[Long, ArrayList[Long]] = this.worker.linkManager.links;
        val places:Set[Long] = linkMap.keySet();
        for (victimId:Long in places)
        {
            val tans:ArrayList[Long] = linkMap.get(victimId); 
            if (!tans.isEmpty())
            {
                countOpenLinks = countOpenLinks + tans.size();
                
                val thief:Long = this.worker.backForth.getForth();
                val victim:Place = Place(victimId);
                val st:PlaceLocalHandle[Worker[Queue, R]] = this.worker.st;

                AssertPrinter.println(  "[Receiver " + here 
                                      + "]: sending delOpen to " + victim
                                      + ".");

                at (victim) @Uncounted async
                {
                    for (val tan:Long in tans)
                    {
                        st().delOpen(tan, thief);
                    }
                }

                tans.clear();

                AssertPrinter.println(  "[Reciever " + here + ":] links to "
                                      + Place(victimId) + ": "
                                      + linkMap.get(victimId).size());
            }
        }
        this.thieves = data.thieves;
        if (null == this.thieves)
        {
            this.thieves = new ArrayList[Long]();
        }
        return countOpenLinks;
    }

//     /**
//      * Interpret the data as incremental update.
//      * Role: backup
//      * sink: yes
//      */
//     @Inline public def incrementalUpdate(val data:BackupData[Queue, R])
//     {
//         assert (null != this.backup) : here + ": UBERPANIC!";
//         assert (null != data) : here + ": PANIC!";
//         if (0 == this.backup.size())
//         {
//             assert (0 == data.removeBottom);
//             this.backup = data.backup;
//         }
//         else
//         {
//             assert (data.removeTop <= this.backup.size())
//                    :   here + "cannot remove " + data.removeTop
//                      + "elements, have only " + this.backup.size() + ".";
// 
//             AssertPrinter.println(  "[Receiver " + here + "]: removeTop = "
//                                   + data.removeTop + ", removeBottom = "
//                                   + data.removeBottom + ", dataSize = "
//                                   + (null == data.backup ?  "null" : data.backup.size()));
// 
//             this.backup.removeFromTop(data.removeTop);
//             if (null != data.backup)
//             {
//                 if (0 < this.backup.size())
//                 {
//                     data.backup.mergeAtBottom(this.backup);
//                 }
//                 this.backup = data.backup;
//             }
//             this.backup.removeFromBottom(data.removeBottom);
//         }
//         // this.checkThieves(data);
//     }

    /**
     * Send the propper respond according to the cause.
     */
    @Inline public def respond(val data:BackupData[Queue, R]) : void
    {
        val tan:Long = data.tan;
        val me:Long = this.worker.me;
        val cause:Long = data.cause;
        val src:Place = Place(data.sourceId);
        val st:PlaceLocalHandle[Worker[Queue, R]] = this.worker.st;
        if (BackupData.CAUSE_STEAL == data.cause)
        {
            AssertPrinter.println(  "[Receiver " + here + ":] sending stlAck "
                                  + "to " + src + ".");

            this.worker.state.valid = false;
            at (src) async
            {
                st().stlAck(tan, me);
            }
        }
        else if (BackupData.CAUSE_REGULAR == data.cause)
        {
            // mb: for debugging
            // AssertPrinter.println(  "[Receiver " + here + ":] backup hash is "
            //                       + this.backup.hash() + ".");

            AssertPrinter.println(  "[Receiver " + here + ":] sending regAck "
                                  + "to " + src + ".");
            at (src) @Uncounted async
            {
                st().regAck(me);
            }
        }
        else if (BackupData.CAUSE_TAKEOVER == data.cause)
        {
            AssertPrinter.println(  "[Receiver " + here + ":] sending toAck "
                                  + "to " + src + ".");
            at (src) async
            {
                st().toAck(me);
            }
        }
        else if (BackupData.CAUSE_IA == data.cause)
        {
            AssertPrinter.println(  "[Receiver " + here + ":] sending iaAck "
                                  + "to " + src + ".");
            this.worker.state.emergency = false;
            at (src) @Uncounted async
            {
                st().iaAck(me);
            }
        }

        if (!this.worker.state.active && !this.worker.state.ghost)
        {
            this.worker.state.ghost = true;
            this.worker.ghostActivity();
        }
    }

    @Inline public def bvEnd() : void
    {
        this.tan = Long.MAX_VALUE;
        this.thieves.clear();
    }
}