package de.uks.plm.ftglb.actor.managers;

import x10.compiler.Inline;
import x10.compiler.Uncounted;

import x10.util.ArrayList;
import x10.util.List;
import x10.util.Pair;

import de.uks.plm.ftglb.actor.TaskBag;
import de.uks.plm.ftglb.actor.TaskQueue;
import de.uks.plm.ftglb.actor.Worker;

import de.uks.plm.ftglb.actor.dataStructures.BackupData;

import de.uks.plm.utils.AssertPrinter;

import de.uks.plm.ftglb.actor.managers.LogManager;
import de.uks.plm.ftglb.actor.logEvents.*;

/**
 * This class manages the sending- and protocolling-process of backup-data.
 * For sending, a reference to the corresponding
 * @link{de.uts.plm.ftglb.actor.Worker worker} is needed.
 */
public class BackupSenderManager[Queue, R] {Queue <: TaskQueue[Queue, R]}
{
    /**
     * Reference to the corresponding
     * @link{de.uts.plm.ftglb.actor.Worker worker}.
     */
    public val worker:Worker[Queue, R];

//     /**
//      * The uppermost task of the min-state. Before sending, all lower tasks
//      * are appended to this queue. Since this is a queue, the result at this
//      * point in time is stored as well.
//      */
//     private var minState:Queue;
// 
//     /**
//      * The size of the stored min-state. Note that this is the size without
//      * the uppermost task.
//      * @link{#minState minState}
//      * @link{de.uts.plm.ftglb.actor.TaskQueue#size() .size()}.
//      */
//     private var minSize:Long; 
// 
//     /**
//      * This attribute hold the size of the last written backup.
//      */
//     private var backupSize:Long;

    public var log:Long;

    public def this(val worker:Worker[Queue, R]) : BackupSenderManager[Queue, R]
    {
        this.worker = worker;
        // this.minSize = 0;
        // this.backupSize = 0;
        // this.minState = (null as Queue);
        this.log = this.worker.me;
    }

//     /**
//      * Protocol the current size of the queue as newS as well as the uppermost
//      * task as theATask.
//      * caller: local only.
//      * 
//      * active while in emergency: no.
//      */
//     @Inline public def prot() : void
//     {
//         if (0 == this.worker.queue.size())
//         {
//             this.minState = this.worker.queue.clone();
//             this.minSize = -1;
//         }
//         else
//         {
//             this.minState = this.worker.queue.getTopElement();
//             this.minSize = this.worker.queue.size() - 1;
//         }
//         // AssertPrinter.println(  "[Sender " + here + "]: new min-state of size "
//         //                       + this.minSize + ".");
//     }
// 
//     /**
//      * Construct the complete queue to send as backup based on the current min-
//      * state.
//      * Sink: yes
//      */
//     @Inline public def constructQueue() : Queue
//     {
//         return (this.constructQueue(0));
//     }

    /**
     * Construct the complete queue to send as backup based on the current min-
     * state, as well as some tasks to remove from bottom (which got stolen 
     * away)
     * Sink: yes
     */
    @Inline public def constructQueue () : Queue
    {
        // assert (this.minState.size() <= 1) : here + ": minState has size > 1!";
        // val queueToSend:Queue = this.minState.clone();
        // if (0 != this.worker.queue.size())
        // {
        //     val rest:Queue = this.worker.queue.getQueueFromBottom(this.minSize);
        //     queueToSend.mergeAtBottom(rest);
        // }

        // assert (queueToSend.size() == this.minSize + 1)
        //        : here + ": constructed queue has wrong size!";

        return (this.worker.queue);
    }

//     /**
//      * This method constructs a
//      * @link{de.uts.plm.ftglb.actor.dataStructures.BackupData
//      * BackupData[Queue, R]}-object.
//      *
//      * @return the @link{de.uts.plm.ftglb.actor.dataStructures.BackupData
//      * BackupData[Queue, R]}-object corresponding to @link{#minState minState}
//      * and @link{#minSize minSize}.
//      */
//     @Inline public def getIncrementalData(val backupId:Long)
//         : BackupData[Queue, R]
//     {
//         val me:Long = this.worker.me;
//         var popTop:Long = this.worker.queue.size() - minSize - 1;
//         val rest:Queue = this.worker.queue.getQueueFromTop(this.minSize);
//         val queueToSend = this.minState.clone();
//         queueToSend.mergeAtBottom(rest);
// 
//         val data:BackupData[Queue, R]
//             = new BackupData[Queue, R](me, popTop, queueToSend);
// 
//         return (data);
//     }
// 
//     /**
//      * Get the size of the minimal protocolled state
//      */
//     @Inline public def minSize() : Long
//     {
//         return (this.minSize);
//     }

    /**
     * Send the backup-data to back.
     * Sink: no, we send the data to back.
     */
    @Inline private def send(val backupId:Long, val data:BackupData[Queue, R])
        : void
    {
        /*send
         * Encapsulate the st in a local val, so not the whole worker-object
         * gets send over the network.
         */
        AssertPrinter.println("[Sender " + here + "]: KEKSE");
        val st:PlaceLocalHandle[Worker[Queue, R]] = this.worker.st;

        AssertPrinter.println(  "[Sender " + here + "]: sending backup to " 
                              + backupId
                              + ", backupSize = " + (null != data.backup ? data.backup.size() : 0)
                              + ", result = " + (null != data.backup
                                                 ? data.backup.getResult().getResult()
                                                 : "oldResult"));

        val backup:Place = Place(backupId);
        at (backup) async
        {
            st().receiver.update(data);
        }
    }
//     /**
//      * Create minimal recorded
//      * @link{ftglb.actor.TaskQueue queue}, pack it into an
//      * @link{ftglb.actor.dataStructures.BackupData 
//      * BackupData[Queue, R]}-object and send it to the place, which backs
//      * this place up.
//      */
//     @Inline public def sendIncrementalBackupRegular(val backupId:Long) : void
//     {
//         var leSendIncrementalBackupRegular:LEsendIncrementalBackupRegular
//             = new LEsendIncrementalBackupRegular(System.nanoTime());
// 
//         worker.logManager.startLE(leSendIncrementalBackupRegular);
//         leSendIncrementalBackupRegular.backupId = backupId;
//         
//         if (this.worker.me == backupId)
//         {
//             this.worker.regAck(backupId);
//             worker.logManager.endLE();
//             return;
//         }
//         val removeTop:Long;
// 
//         val queueToSend:Queue;
//         if (this.minSize + 1  >= this.backupSize)
//         {
//             removeTop = 1;
//             queueToSend = this.constructQueue();
// 
//             assert (queueToSend.size() == this.minSize + 1)
//                    :  here + ": backup has wrong size of " + queueToSend.size()
//                     + " (should have " + (this.minSize + 1) + ").";
// 
//             val stableTasks:Long = Math.max(this.backupSize - 1, 0);
//             queueToSend.removeFromBottom(stableTasks);
// 
//             assert (queueToSend.size() + stableTasks == this.minSize + 1)
//                     : here + ": after removeBottom, the size of sending queue is wrong.";
//         }
//         else // if (this.minSize + 1 < this.backupSize
//         {
//             removeTop = Math.min(this.backupSize - this.minSize
//                                  , this.backupSize);
// 
//             queueToSend = this.minState.clone();
//         }
// 
//         val me = this.worker.me;
// 
//         //Ansatzpunkt zum holen der Daten für LogManager.
//         val data:BackupData[Queue, R]
//             = new BackupData[Queue, R](me, removeTop, queueToSend);
// 
//         AssertPrinter.println(  here + ": SendingIncrementalBackup, size "
//                               + queueToSend.size());
// 
//         this.send(backupId, data);
//         this.backupSize = this.minSize + 1;
// 
//         AssertPrinter.println(  "[Sender " + here + "]: backupSize is now "
//                               + this.backupSize);
// 
//         this.worker.resetK();
// 
//         worker.logManager.endLE();
//     }

    /**
     * This method constructs and sends a full backup. Note, that the full
     * @link{de.uts.ftglb.actor.Worker#queue worker.queue} at this point
     * in time is recorded in @link{#minState} and @link{minsize} and sent to
     * the backup-place.
     */
    @Inline public def sendFullBackupRegular(val backupId:Long) : void
    {
        var leSendFullBackupRegular:LEsendFullBackupRegular
            = new LEsendFullBackupRegular(System.nanoTime());
        worker.logManager.startLE(leSendFullBackupRegular);
        leSendFullBackupRegular.backupId = backupId;
        
        if (this.worker.me == backupId)
        {
            this.worker.regAck(backupId);
            this.worker.state.afterMerge = false;
            this.worker.resetK();
            worker.logManager.endLE();
            return;
        }
        val queueToSend:Queue = this.constructQueue();
        val data:BackupData[Queue, R] 
            = new BackupData[Queue, R](this.worker.me, queueToSend);

        AssertPrinter.println("[Sender " + here + "]: sending full, regular backup.");
        this.send(backupId, data);
        // this.backupSize = this.minSize + 1;
        //leSendFullBackupRegular.setBackupSize(this.backupSize);

        AssertPrinter.println(  "[Sender " + here + "]: backupSize is now "
                              + queueToSend.size());

        // mb: for debugging
        // AssertPrinter.println(  "[Sender " + here + "]: hash is "
        //                       + queueToSend.hash());

        this.worker.state.afterMerge = false;
        this.worker.resetK();

        worker.logManager.endLE();
    }

    /**
     * Send a full steal-backup
     * sink: no, indirectly switching places via send.
     */
    @Inline private def sendFullBackupSteal(val tan:Long
                                           , val backupId:Long
                                           , val removeBottom:Long
                                           , val thieves:List[Long]) : void
    {
        if (this.worker.me == backupId)
        {
            this.worker.stlAck(tan, backupId);
            return;
        }

        val queueToSend:Queue = this.constructQueue();

        AssertPrinter.println(  here + ": currentQueue-size: " + queueToSend.size() + ".");

        AssertPrinter.println(  here + ": sending full-steal backup of size "
                              + this.worker.queue.size() + ".");

        val data:BackupData[Queue, R]
            = new BackupData[Queue, R](tan
                                       , this.worker.me
                                       , queueToSend
                                       , thieves);

        this.send(backupId, data);

//         this.backupSize = this.minSize + 1;
// 
//         AssertPrinter.println(  "[Sender " + here + "]: backupSize is now "
//                               + this.backupSize);
    }

    /**
     * This method acts for STLNreq and STLTreq. Depending on stolen, the
     * Backup-manager deciedes, which kind of backup to send and constructs
     * the backup-data accordingly.
     *
     * Important: this method DOES NOT CHECK, whether we can send. This must be
     * done before calling this method.
     *
     * TODO: maybe we can simplify this.
     */
    @Inline public def sendStealBackup(val tan:Long
                                       , val backupId:Long
                                       , val stolen:Long
                                       , val thieves:List[Long]) : void
    {
        var leSendStealBackup:LEsendStealBackup
            = new LEsendStealBackup(System.nanoTime());
        worker.logManager.startLE(leSendStealBackup);
        leSendStealBackup.backupId = backupId;
        
        if (this.worker.me == backupId)
        {
            this.worker.stlAck(tan, backupId);
            worker.logManager.endLE();
            return;
        }
//         /**
//          * 1st case: we have not merged anything since the last backup and the
//          * amount of stolen tasks is smaller than the protocolled min-state and
//          * smaller than the last backup -> Simply send the number of stolen
//          * tasks to the backup place. Do not reset the cycle.
//          */
//         if (   (stolen <= backupSize)
//             && (stolen <= this.minSize)
//             && !this.worker.state.afterMerge)
//         {
//             AssertPrinter.println("[Sender " + here + "]: case 1");
// 
//             val backup:BackupData[Queue,R]
//                 = new BackupData[Queue, R](tan
//                                            , this.worker.me
//                                            , BackupData.KIND_INCR
//                                            , BackupData.CAUSE_STEAL
//                                            , 0
//                                            , stolen
//                                            , -1
//                                            , (null as Queue)
//                                            , thieves);
// 
//             AssertPrinter.println("[Sender " + here + "]: constructed backup in case 1");
//             AssertPrinter.println("[Sender " + here + "]: calling Send in case 1");
// 
//             this.send(backupId, backup);
//             this.backupSize -= stolen;
//             this.minSize -= stolen;
// 
//             AssertPrinter.println(  "[Sender " + here + "]: backupSize is now "
//                                   + this.backupSize);
//         }
//         /**
//          * 2nd cast: no merge since the last backup and the number of stolen 
//          * tasks is bigger than the last backup, but smaller than the current 
//          * min-state -> send the current min-state without the stolen tasks, but
//          * do NOT reset the cycle. Basically a full backup.
//          * 5-a)th case: We had a merge after the last backup, but the amount of
//          * stolen tasks is smaller than the protocolled mind-state -> same
//          * reaction.
//          */
//         else if (   (stolen > backupSize && stolen <= this.minSize)
//                  || (this.worker.state.afterMerge && (stolen <= this.minSize)))
//         {
//             AssertPrinter.println("[Sender " + here + "]: case 2");
// 
//             this.minSize -= stolen;
// 
//             this.sendFullBackupSteal(tan, backupId, stolen, thieves);
//         }
        /**
         * 3rd and 4th case: no merge since the last backup, but the number
         * of stolen tasks is bigger than the current min-state -> write 
         * full backup, reset cycle.
         * 5-bth case: merge since last backup and and the number of stolen
         * tasks is bigger than current min-state -> write full backup, reset
         * cycle
         */
        // else // if (stolen > this.minSize)
        // {
            AssertPrinter.println("[Sender " + here + "]: case 3");
            // this.prot();

            this.sendFullBackupSteal(tan, backupId, 0, thieves);
            this.worker.resetK();
        // }
        worker.logManager.endLE();
    }

    /**
     * Send a full backup
     * sink: no, indirectly via send
     */
    @Inline private def sendFullBackup(val backupId:Long, val cause:Long) : void
    {
        var tan:Long = Long.MAX_VALUE;
        var thieves:List[Long] = null;
        if (this.worker.state.repeatStlReq)
        {
            tan = this.worker.currentTan;
            thieves = new ArrayList[Long]();
            for (entry:Pair[Long, TaskBag] in this.worker.tanDescriptor)
            {
                thieves.add(entry.first);
            }
        }

        val backup:BackupData[Queue, R]
            = new BackupData[Queue, R] (tan
                                        , this.worker.me
                                        , BackupData.KIND_COMP
                                        , cause
                                        // , 0
                                        // , 0
                                        , this.log
                                        , this.worker.queue
                                        , thieves);

        this.send(backupId, backup);
        // this.backupSize = this.minSize + 1;

        AssertPrinter.println(  "[Sender " + here + "]: backupSize is now "
                              + this.worker.queue.size());
    }

    /**
     * This method functions as IA-Backup.
     * sink: no, indirectly via send
     */
    @Inline public def sendFullBackupIA(val backupId:Long) : void
    {
        var leSendFullBackupIA:LEsendFullBackupIA
            = new LEsendFullBackupIA(System.nanoTime());
        worker.logManager.startLE(leSendFullBackupIA);
        leSendFullBackupIA.backupId = backupId;
        
        if (this.worker.me == backupId)
        {
            this.worker.toAck(backupId);
            worker.logManager.endLE();
            return;
        }
        this.sendFullBackup(backupId, BackupData.CAUSE_IA);

        worker.logManager.endLE();
    }

    /**
     * This method functions as TO-Backup
     * sink: no, indirectly via send
     */
    @Inline public def sendFullBackupTO(val backupId:Long)
    {
        var leSendFullBackupTO:LEsendFullBackupTO
            = new LEsendFullBackupTO(System.nanoTime());
        worker.logManager.startLE(leSendFullBackupTO);
        leSendFullBackupTO.backupId = backupId;
        
        if (this.worker.me == backupId)
        {
            this.worker.toAck(backupId);
            worker.logManager.endLE();
            return;
        }
        val tan:Long = Math.min(this.worker.currentTan, Long.MAX_VALUE);
        var thieves:List[Long] = new ArrayList[Long]();
        val tanDescriptor:List[Pair[Long, TaskBag]] = this.worker.tanDescriptor;
        for (entry:Pair[Long, TaskBag] in tanDescriptor)
        {
            thieves.add(entry.first);
        }
        if (thieves.isEmpty())
        {
            thieves = null;
        }

        val backup:BackupData[Queue, R]
            = new BackupData[Queue, R](tan
                                       , this.worker.me
                                       , BackupData.KIND_COMP
                                       , BackupData.CAUSE_TAKEOVER
                                       // , 0
                                       // , 0
                                       , -1
                                       , this.worker.queue
                                       , thieves);

        AssertPrinter.println(  "[Sender " + here +  "]: sending TO-Backup to "
                              + Place(backupId) + ". Thieves = " + thieves);

        this.send(backupId, backup);
        // this.backupSize = this.minSize + 1;

        AssertPrinter.println(  "[Sender " + here + "]: backupSize is now "
                              + this.worker.queue.size());

        worker.logManager.endLE();
    }
}