package de.uks.plm.ftglb.actor.managers;

import de.uks.plm.ftglb.actor.logEvents.*;
import x10.io.File;
import x10.compiler.Inline;
import de.uks.plm.ftglb.actor.managers.ProcessingStackManager;
import de.uks.plm.ftglb.actor.managers.WaitingAndWorkingManager;
import x10.util.ArrayList;

public final class LogManager
{
  public static val WORKING:Long = 0;
  public static val IDLING:Long = 1;
  public static val STEAL_RECEIVING:Long = 2;
  public static val TRY_STEAL:Long = 3;
  public static val PROCESS_STEALS:Long = 4;
  public static val DISTRIBUTE:Long = 5;
  public static val GIVE:Long = 6;
  public static val GIVE_LOOT_TO_THIEVES:Long = 7;
  public static val VICTIM_LINK:Long = 8;
  public static val PROCESS_MERGES:Long = 9;
  public static val RECEIVE_BACKUP:Long = 10;
  public static val SEND_FULL_BACKUP_REGULAR:Long = 11;
  public static val SEND_INCREMENTAL_BACKUP_REGULAR:Long = 12;
  public static val SEND_STEAL_BACKUP:Long = 13;
  public static val SEND_FULL_BACKUP_IA:Long = 14;
  public static val SEND_FULL_BACKUP_TO:Long = 15;
  public static val REG_ACK:Long = 16;
  public static val TO_ACK:Long = 17;
  public static val IA_ACK:Long = 18;
  public static val RST_ACK:Long = 19;
  public static val BV_END:Long = 20;
  public static val F_END:Long = 21;
  public static val STL_ACK:Long = 22;
  public static val DEL_OPEN:Long = 23;
  public static val BF_ACK:Long = 24;
  public static val CHECK_TIMEOUTS:Long = 25;
  public static val NO_TASKS:Long = 26;
  public static val STEAL_REJECT:Long = 27;

  /** Übergibt die Laufzeit von processStack an Working. **/
  public var processingStackManager:ProcessingStackManager;
  /** Misst und übergibt die Laufzeit von Distributing und 
   '* wenn waiting=true ist während Working läuft. */
  public var waitingAndWorkingManager:WaitingAndWorkingManager;
  /** Id des Places zu dem das Objekt des LogManagers gehört. */
  private val placeId:Long;
  /** Referenzzeit des LogManagers. Diese ist mit den LogManager-Objekte der
    * anderen Places synchronisiert */
  private val refTime:Long;
  /** Initiales LogEvent; Dient als Startpunkt der doppelt verketteten List von
    * LogEvents. */
  private val initLE:LogEvent;
  /** Momentaner Endpunkt der doppelt verketteten Liste von LogEvents. */
  public var currLE:LogEvent; //public vor startWaiting(), endWaiting calls in Worker.x10.

  /** Verhältnis zwischen startLE und endLE; Zum debugging. */
  private var startEndBalance:Long;

  /** Länge der Kette von bereits angelegten LogEvents; Zum debugging. */
  private var listLength:Long;
  /** Konstruktor des LogManagers. Wird in der Initialisierungphase der GLB auf
    * allen Places aufgerufen. Die Referenzzeit wird vom GLB Logger
    * übernommen.*/

  private var numLE:Long;
  private var listLabel:Rail[String];
  private var listConstructor:Rail[(Long) => LogEvent];
  private var listDoc:Rail[String];

  //private var state:StateManager;

  /*public def this(placeId:Long, refTime:Long)
  {
    this(placeId, refTime);
    //this.state = state;
  }*/

  public def this(placeId:Long, refTime:Long, psm:ProcessingStackManager, 
    wawm:WaitingAndWorkingManager)
  {
    this.processingStackManager = psm;
    this.waitingAndWorkingManager = wawm;

    this.placeId = placeId;
    this.refTime = refTime;
    Console.OUT.println("Initialization of LogManager on Place "
      + this.placeId + " with reference time "
      + this.refTime + " ns.");

    this.numLE = 28;
    this.listLabel = new Rail[String](numLE);
    this.listConstructor = new Rail[(Long) => LogEvent](numLE);
    this.listDoc = new Rail[String](numLE);

    this.listLabel(WORKING)  = "working";
    this.listLabel(IDLING)  = "idling";
    this.listLabel(2)  = "stealReceiving";
    this.listLabel(3)  = "trySteal";
    this.listLabel(4)  = "processSteals";
    this.listLabel(DISTRIBUTE)  = "distribute";
    this.listLabel(GIVE)  = "give";
    this.listLabel(7)  = "giveLootToThieves";
    this.listLabel(8)  = "victimLink";
    this.listLabel(9)  = "processMerges";
    this.listLabel(RECEIVE_BACKUP) = "receiveBackup";
    this.listLabel(11) = "sendFullBackupRegular";
    this.listLabel(12) = "sendIncrementalBackupRegular";
    this.listLabel(13) = "sendStealBackup";
    this.listLabel(14) = "sendFullBackupIA";
    this.listLabel(15) = "sendFullBackupTO";
    this.listLabel(16) = "regAck";
    this.listLabel(17) = "toAck";
    this.listLabel(18) = "iaAck";
    this.listLabel(19) = "rstAck";
    this.listLabel(20) = "bvEnd";
    this.listLabel(21) = "fEnd";
    this.listLabel(22) = "stlAck";
    this.listLabel(23) = "delOpen";
    this.listLabel(24) = "bfAck";
    this.listLabel(25) = "checkTimeouts";
    this.listLabel(26) = "noTasks";
    this.listLabel(STEAL_REJECT) = "stealReject";

    this.listConstructor(WORKING)  = (sTime:Long) => {return new LEworking(sTime);};
    this.listConstructor(IDLING)  = (sTime:Long) => {return new LEidling(sTime);};
    this.listConstructor(2)  = (sTime:Long) => {return new LEstealReceiving(sTime);};
    this.listConstructor(3)  = (sTime:Long) => {return new LEtrySteal(sTime);};
    this.listConstructor(4)  = (sTime:Long) => {return new LEprocessSteals(sTime);};
    this.listConstructor(DISTRIBUTE)  = (sTime:Long) => {return new LEdistribute(sTime);};
    this.listConstructor(GIVE)  = (sTime:Long) => {return new LEgive(sTime);};
    this.listConstructor(7)  = (sTime:Long) => {return new LEgiveLootToThieves(sTime);};
    this.listConstructor(8)  = (sTime:Long) => {return new LEvictimLink(sTime);};
    this.listConstructor(9)  = (sTime:Long) => {return new LEprocessMerges(sTime);};
    this.listConstructor(RECEIVE_BACKUP) = (sTime:Long) => {return new LEreceiveBackup(sTime);};
    this.listConstructor(11) = (sTime:Long) => {return new LEsendFullBackupRegular(sTime);};
    this.listConstructor(12) = (sTime:Long) => {return new LEsendIncrementalBackupRegular(sTime);};
    this.listConstructor(13) = (sTime:Long) => {return new LEsendStealBackup(sTime);};
    this.listConstructor(14) = (sTime:Long) => {return new LEsendFullBackupIA(sTime);};
    this.listConstructor(15) = (sTime:Long) => {return new LEsendFullBackupTO(sTime);};
    this.listConstructor(16) = (sTime:Long) => {return new LEregAck(sTime);};
    this.listConstructor(17) = (sTime:Long) => {return new LEtoAck(sTime);};
    this.listConstructor(18) = (sTime:Long) => {return new LEiaAck(sTime);};
    this.listConstructor(19) = (sTime:Long) => {return new LErstAck(sTime);};
    this.listConstructor(20) = (sTime:Long) => {return new LEbvEnd(sTime);};
    this.listConstructor(21) = (sTime:Long) => {return new LEfEnd(sTime);};
    this.listConstructor(22) = (sTime:Long) => {return new LEstlAck(sTime);};
    this.listConstructor(23) = (sTime:Long) => {return new LEdelOpen(sTime);};
    this.listConstructor(24) = (sTime:Long) => {return new LEbfAck(sTime);};
    this.listConstructor(25) = (sTime:Long) => {return new LEcheckTimeouts(sTime);};
    this.listConstructor(26) = (sTime:Long) => {return new LEnoTasks(sTime);};
    this.listConstructor(STEAL_REJECT) = (sTime:Long) => {return new LEstealReject(sTime);};

    this.listDoc(WORKING)  = "sTime wTime eTime working predecessor successor processing waiting distributing";
    this.listDoc(IDLING)  = "sTime wTime eTime idling predecessor successor";
    this.listDoc(2)  = "sTime wTime eTime stealReceiving predecessor successor thiefId random buddy onThieves onLifelineThieves";
    this.listDoc(3)  = "sTime wTime eTime trySteal predecessor successor victimId stealingType wasSent sendingTime";
    this.listDoc(4)  = "sTime wTime eTime processSteals predecessor successor";
    this.listDoc(DISTRIBUTE)  = "sTime wTime eTime distribute predecessor successor";
    this.listDoc(GIVE)  = "sTime wTime eTime give predecessor successor tan victimId lootSize isRecorded merging";
    this.listDoc(7)  = "sTime wTime eTime giveLootToThieves predecessor successor tan tanDescriptorSize tanThiefIds lootSize";
    this.listDoc(8)  = "sTime wTime eTime victimLink predecessor successor";
    this.listDoc(9)  = "sTime wTime eTime processMerges predecessor successor";
    this.listDoc(RECEIVE_BACKUP) = "sTime wTime eTime receiveBackup predecessor successor kindBack causeBack sourceId openLinks";
    this.listDoc(11) = "sTime wTime eTime sendFullBackupRegular predecessor successor backupId";
    this.listDoc(12) = "sTime wTime eTime sendIncrementalBackupRegular predecessor successor backupId";
    this.listDoc(13) = "sTime wTime eTime sendStealBackup predecessor successor backupId";
    this.listDoc(14) = "sTime wTime eTime sendFullBackupIA predecessor successor backupId";
    this.listDoc(15) = "sTime wTime eTime sendFullBackupTO predecessor successor backupId";
    this.listDoc(16) = "sTime wTime eTime regAck predecessor successor srcId";
    this.listDoc(17) = "sTime wTime eTime toAck predecessor successor srcId";
    this.listDoc(18) = "sTime wTime eTime iaAck predecessor successor srcId";
    this.listDoc(19) = "sTime wTime eTime rstAck predecessor successor srcId";
    this.listDoc(20) = "sTime wTime eTime bvEnd predecessor successor srcId";
    this.listDoc(21) = "sTime wTime eTime fEnd predecessor successor srcId tan";
    this.listDoc(22) = "sTime wTime eTime stlAck predecessor successor srcId tan";
    this.listDoc(23) = "sTime wTime eTime delOpen predecessor successor placeId tan";
    this.listDoc(24) = "sTime wTime eTime bfAck predecessor successor thiefId tan contactThief";   
    this.listDoc(25) = "sTime wTime eTime checkTimeouts predecessor successor waitingForSize checksFailed";
    this.listDoc(26)  = "sTime wTime eTime noTasks predecessor successor thiefId stealType";
    this.listDoc(STEAL_REJECT) = "sTime wTime eTime stealReject predecessor successor victimId rejectType";

    /** Initiales LogEvent ist Idling. */
    //this.initLE = eventFactory("idling", this.refTime);
    this.initLE = eventFactory(this.IDLING, this.refTime);
    this.initLE.prevLE = this.initLE;
    this.initLE.firstPartLE = this.initLE;
    this.initLE.predecessor = 0;
    this.currLE = this.initLE;
    this.initLE.wTime = System.nanoTime();

    this.startEndBalance = 0;
    this.listLength = 0;
  }

  /** Startet ein neues LogEvent und unterbricht gleichzeitig das aktuell laufende
    * LogEvent. Jeder Aufruf von startLE muss durch endLE beendet werden.*/

  public def getRefTime() : Long
  {
    return this.refTime;
  }  

  //Referenze auf das neue LE muss zurück gegeben werden, damit Felder setzen kann.
  public def startLE(label:Long, sTime:Long) : void
  {
    startLE(eventFactory(label, sTime));
  }

  public def startLE(newLE:LogEvent) : void
  {
    // Ich habe currLE und newLE.
    // Beende currLE.      
    this.currLE.eTime = newLE.sTime;
    if(this.currLE.label.equals("working"))
    {
      // Working wird beendet.
      this.waitingAndWorkingManager.endWorking(this.currLE.eTime);
      this.processingStackManager.endWorking();
    }
    if(newLE.label.equals("working"))
    {
      // Working wird gestartet.
      this.waitingAndWorkingManager.startWorking(newLE.sTime);
      this.processingStackManager.startWorking();
    }
    /*if (state.waiting)
    {
      //Console.OUT.println("Place " + placeId + " : Label " + this.currLE.label + " : state.waiting " + state.waiting);
    }
    if (this.currLE.label.equals("idling") && state.waiting)
    {
      Console.OUT.println("Debug: Call endWaiting in startLE.");
      this.currLE.endWaiting(this.currLE.eTime);
    }
    if (newLE.label.equals("idling") && state.waiting)
    {
      Console.OUT.println("Debug: Call startWaiting in startLE.");
      newLE.startWaiting(newLE.sTime);
    }*/
    // Setze Nachfolger von currLE.
    this.currLE.nextLE = newLE;
    // Setze Vorgänger von newLE.
    newLE.prevLE = this.currLE;

    // Mit startLE werden neue LE gestartet. Setzte Vorgänger 0 und fistPartLE auf sich selbst.
    newLE.predecessor = 0;
    newLE.firstPartLE = newLE;

    // Setze das Ende der doppelt verketteten Liste auf newLE.
    newLE.wTime = System.nanoTime();
    this.currLE = newLE; 
    /** newLE hat 8 zu setzende Felder. 5 davon werden hier gesetzt: sTime, prevLE,
      * predecessor, firstPartLE, wTime.
      * Die Felder nextLE, sucessor und eTime können noch nicht gesetzt werden.
      */
    /** Von currLE werden eTime und nextLE gesetzt.
      */
    /** Wird zweimal hintereinander startLE aufgerfen bleibt von currLE sucessor
      * offen. Das ist ok, denn sucessor wird erst gesetzt wenn das
      * unterbrochene LE wieder fortgesetzt wird.*/

    startEndBalance = startEndBalance + 1;
    listLength = listLength + 1;
    if (0 == listLength%10000)
    {
      Console.OUT.println("Auf Place " + here + " sind " + listLength + " LE");
    }
  }

  /** Beendet ein LogEvent, bestimmt den Vorgänger des ersten Teils des Vorgänger
    * LogEvents und setzt diesen fort. */
  public def endLE() : void
  {
    // Startzeit für newLE um den Overhead zu erfassen.
    var sTimeNewLE:Long = System.nanoTime();
    // Erster Teil von currLE.
    var firstPartCurrLE:LogEvent = this.currLE.firstPartLE;
    // Setze predecessor currLE.
    // Setze successor firstPartCurrLE.
    // Setze successor currLE auf 0. endLE Beendet LE.
    this.currLE.successor = 0;
    // Vorgänger des ersten Teils von currLE soll als newLE fortgesetzt werden.
    var resumeLE:LogEvent = firstPartCurrLE.prevLE;
    var resumeEventLabel:String = resumeLE.label;
    // Erzeuge newLE.
    var newLE:LogEvent = eventFactory(resumeEventLabel, sTimeNewLE);
    // Setze Startzeit von newLE;
    //newLE.sTime = sTimeNewLE; //Wird schon im Konstruktor gemacht.
    // Beende currLE.
    //this.currLE.eTime = sTimeNewLE;
    this.currLE.eTime = newLE.sTime;
    if(this.currLE.label.equals("working"))
    {
      // Working wird beendet.
      this.waitingAndWorkingManager.endWorking(this.currLE.eTime);
      this.processingStackManager.endWorking();
    }
    if(newLE.label.equals("working"))
    {
      // Working wird gestartet.
      this.waitingAndWorkingManager.startWorking(newLE.sTime);
      this.processingStackManager.startWorking();
    }
    /*if (this.currLE.label.equals("idling") && state.waiting) 
    { 
      Console.OUT.println("Debug: Call endWaiting in endLE.");
      this.currLE.endWaiting(this.currLE.eTime);
    }
    if (newLE.label.equals("idling") && state.waiting) 
    {
      Console.OUT.println("Debug: Call startWaiting in endLE.");
      this.currLE.startWaiting(this.currLE.sTime);
    }*/
    // Setze Nachfolger von currLE;
    this.currLE.nextLE = newLE;
    // Setze Vorgänger von newLE;
    newLE.prevLE = this.currLE;
    // Bestimme predecessor newLE.
    var n:Long = countToEnd(resumeLE);
    // Setze predecessor newLE.
    newLE.predecessor = n;
    // Setze successor resumeLE.
    resumeLE.successor = n;
    // Setze firstPartLE newLE.
    newLE.firstPartLE = resumeLE.firstPartLE;
    // Setze neues currLE.
    this.currLE = newLE;
    // Setze wTime zur messung des Overheads.
    this.currLE.wTime = System.nanoTime();

    startEndBalance = startEndBalance -1;
    listLength = listLength + 1;
    if (0 == listLength%10000)
    {
      Console.OUT.println("Auf Place " + here + " sind " + listLength + " LE");
    }
  }

  /** Zählt die Schritte von startLE bis zu currLE. */
  private def countToEnd(startLE:LogEvent) : Long
  {
    var tempLE:LogEvent = startLE;
    var n:Long = 0;
    while (tempLE.nextLE != null)
    {
      n = n + 1;
      tempLE = tempLE.nextLE;
    }
    return n;
  }

  /** Geht in der doppelten Kette von LogEvents n Schritte zurück. */
  /*private def stepBack(currLE:LogEvent, n:Long) : LogEvent
  {
    predecessorLE:LogEvent = currLE;
    for (var i:Long = 0; i < n; i++)
    {
      predecessorLE = predecessorLE.prevLE;
    }
    return predecessorLE;
  }*/

  /*private def eventFactory(label:String) :LogEvent
  {
    return eventFactory(label, System.nanoTime());
  }*/

  /** Generates LogEvent objects */
  private def eventFactory(label:String, sTime:Long) : LogEvent
  {
    //Console.OUT.println(this.numLE);
    for (var i:Long = 0; i < this.numLE; ++i)
    {
      //Console.OUT.println(this.listLabel(i));
      if (label.equals(this.listLabel(i)))
      {
        //Console.OUT.println("Label = " + label);
        return this.listConstructor(i)(sTime);
      }
    }
    //Console.OUT.println("Label_ = " + label);
    Console.OUT.println("Error: Method eventFactory returns null!");
    return null;
  }

  private def eventFactory(label:Long, sTime:Long) :LogEvent
  {
    return this.listConstructor(label)(sTime);
  }

  /** Beendet das aktuell laufende LogEvent, ohne ein neues LogEvent zu
    * starten. */
  public def endLogging() : void
  {
    this.currLE.nextLE = null; //Doppelt gesetzt.
    this.currLE.successor = 0;
    this.currLE.eTime = System.nanoTime();

    //Console.OUT.println("startEndBalance = " + startEndBalance);
    if (0 != startEndBalance)
    {
      Console.OUT.println("Error: startEndBalance not equal zero");
    }
    Console.OUT.println(this.listLength + " LogEvents on Place " + this.placeId + ".");
  }

  /** Gibt die LogEvents für jeden Place in eine Datei aus. Das Logging muss
    * vorher beendet worden sein. */
  public def printToFile() : void
  {
    val outputPath = "out/Logfile" + this.placeId + ".txt";
    val O = new File(outputPath);
    val P = O.printer();

    val p : ArrayList[Long] = this.processingStackManager.getProcessing();
    val w : ArrayList[Long] = this.waitingAndWorkingManager.getWaiting();
    val d : ArrayList[Long] = this.waitingAndWorkingManager.getDistribute();

    var i:Long = 0;

    var printEvent:LogEvent = initLE;
    P.println(printEvent.toString(this.refTime));
    while (printEvent.nextLE != null) {
      printEvent = printEvent.nextLE;
      //printEvent.postprocessing();
      if (printEvent.label.equals("working"))
      {
        // Hänge Felder an den String von Working
        val str : String = printEvent.toString(this.refTime) 
        + " " + p.get(i) + " " + w.get(i) + " " + d.get(i);
        P.println(str);
        // Für jedes Working LE gibt es einen Eintrag.
        i = i + 1;
      } else {
        P.println(printEvent.toString(this.refTime));
      }
    }
    P.flush();
  }

 public def printSLEToFile(str:String) : void
 {
   val outputPath = "out/SimpleLogfile" + this.placeId + ".txt";
   val O = new File(outputPath);
   val P = O.printer();
   P.println(str);
   P.flush();
 }

  public def printListDoc() : void
  {
    if (0 == this.placeId)
    {
      val outputPath = "out/Doc.txt";
      val O = new File(outputPath);
      val P = O.printer();

      for (i:String in listDoc)
      {
      P.println(i);
      }
      P.flush();
    }
  }
}
