package de.uks.plm.ftglb.actor.managers;

import x10.util.ArrayList;
import x10.util.Collection;
import x10.util.List;
import x10.util.Pair;

/**
 * This class manages received tans on the side of the thief, which have not
 * been closed yet (i.e. the thief has received and merged the tasks, but the
 * victim has not yet reacted to this information.
 *
 * @author Marco Bungart
 */
public class OpenFEndManager
{
    /**
     * openFEnd: 1st component ist the tan, 2nd component is the place-id
     */
    public val openFEnd:ArrayList[Pair[Long, Long]];

    /**
     * Default-constructor, setting the initial size of
     * @link{#openFEnd openFEnd} to 16.
     */
    public def this():OpenFEndManager
    {
        this(16);
    }

    /**
     * Constructor.
     *
     * @param initSize the initial size for @link{#openFEnd openFEnd}.
     */
    public def this(val initSize:Long):OpenFEndManager
    {
        this.openFEnd = new ArrayList[Pair[Long, Long]](initSize);
    }

    /**
     * Add tan and place to @link{#openFend openFend}.
     *
     * @param place the placd-id to add.
     * @param tan the tan to add.
     */
    public def add(val tan:Long, val placeId:Long):void
    {
        val entry:Pair[Long, Long] = new Pair[Long, Long](tan, placeId);
        this.openFEnd.add(entry);
    }

    /**
     * Add a @link{x10.util.Collection Collection} of @link{x10.util.Pair Pairs}
     * of tan (@link{x10.lang.Long Long}) and place-id
     * (@link{x10.lang.Long Long}) to @link{#openFend openFend}.
     *
     * @param entries the collection to add.
     */
    private def add(val entries:Collection[Pair[Long, Long]]):void
    {
        this.openFEnd.addAll(entries);
    }

    /**
     * Returns a @link{x10.util.List List} of all place-ids, for which this
     * tan has an entry.
     *
     * @param val tan the tan to search for.
     *
     * @return the list of place-ids.
     */
    public def contains(val tan:Long):List[Long]
    {
        val result:List[Long] = new ArrayList[Long]();
        for (entry:Pair[Long, Long] in this.openFEnd)
        {
            if (tan == entry.first)
            {
                result.add(entry.second);
            }
        }
        return (result);
    }

    /**
     * Removes the specified place and the specified tan from
     * @link{openFEnd openFEnd}.
     *
     * @param place the place-id to remove
     * @param tan the tan to remove.
     *
     * @return true, if a corresponding entry was found,
     *         false if not.
     */
    public def remove(val tan:Long, val place:Long):Boolean
    {
        for (index:Long in 0..(this.openFEnd.size() - 1))
        {
            val entry:Pair[Long, Long] = this.openFEnd(index);

            if (tan == entry.first && place == entry.second)
            {
                this.openFEnd.removeAt(index);
                return true;
            }
        }
        return false;
    }
}
