package de.uks.plm.ftglb.actor.managers;

import x10.util.ArrayList;
import x10.compiler.Inline;
import x10.io.File;

public final class WaitingAndWorkingManager
{
  private static val ARRAY_LENGTH:Long = 10000;

  private var workingIsRunning:Boolean;
  private var waiting:Boolean;
  private var startMeasurement:Long;
  private var endMeasurement:Long;
  private var waitingAndWorking:Long;
  private var startDistributing:Long;
  private var endDistributing:Long;
  private var distributingRuntime:Long;
  private var wawList:ArrayList[Long];
  private var dList:ArrayList[Long];

  public def this()
  {
    this.workingIsRunning = false;
    this.waiting = false;
    this.startMeasurement = -1;
    this.endMeasurement = -1;
    this.waitingAndWorking = 0;
    this.startDistributing = -1;
    this.endDistributing = -1;
    this.distributingRuntime = 0;
    this.wawList = new ArrayList[Long](10000);
    this.dList = new ArrayList[Long](10000);
  }
  /* Wird vom StateManager aufgerufen wenn waiting = true gesetzt wird (es 
   * muss vorher false gewesen sein). */
  @Inline public def startWaiting() : void
  {
    if (this.workingIsRunning == true && this.waiting == false)
    {
      this.startMeasurement = System.nanoTime();
    }
    this.waiting = true;
  }

  /* Wird vom StateManager aufgerufen wenn waiting = false gesetzt wird (es
   * muss vorher false gewesen sein. */
  @Inline public def endWaiting() : void
  {
    if (this.workingIsRunning == true && this.waiting == true)
    {
      this.endMeasurement = System.nanoTime();
      this.calcWaiting();
    }
    this.waiting = false;
  }

  /* Startet die Zeitmessung von Distributing. Da Working nicht unterbrochen 
   * wird wenn die Methode läuft und waiting = false sein muss, ist keine 
   * weitere Absicherung nötig. */
  @Inline public def startDistributing() : void
  {
    this.startDistributing = System.nanoTime();
  }

  @Inline public def endDistributing() : void
  {
    this.endDistributing = System.nanoTime();
    this.calcDistribute();
  }

  /* Wird im LogManager beim Start von Working aufgerufen. Bekommt die sTime
   * von Working übergeben. */
  @Inline public def startWorking(val startTime:Long) : void
  {
    this.workingIsRunning = true;
    this.resetMeasurement();
    if (this.waiting == true)
    {
      this.startMeasurement = startTime;
    }
  }

  /* Wird im LogManager beim Beenden von Working aufgerufen. Gibt die Zeit zeit
   * zurück die waiting = true während der Laufzeit des Events war und 
   * setzt diese zurück. Bekommt die eTime des beendeten working zurück.*/
  @Inline public def endWorking(val endTime:Long) : void
  {
    this.workingIsRunning = false;
    if (this.waiting == true)
    {
      this.endMeasurement = endTime;
      this.calcWaiting();
    }
    // Working wird beendet, die Ergebnisse werden gesepeichert.
    this.addResult();
  }

  @Inline private def calcDistribute() : void
  {
    this.distributingRuntime = this.distributingRuntime
    + (this.endDistributing - this.startDistributing);
  }

  @Inline private def calcWaiting() : void
  {
    this.waitingAndWorking = this.waitingAndWorking
    + (this.endMeasurement - this.startMeasurement);
  }

  @Inline private def addResult() : void
  {
    this.wawList.add(this.waitingAndWorking);
    this.dList.add(this.distributingRuntime);
  }

  @Inline private def resetMeasurement() : void
  {
    this.waitingAndWorking = 0;
    this.distributingRuntime = 0;
  }

  @Inline public def getDistribute() : ArrayList[Long]
  {
    return this.dList;
  }

  @Inline public def getWaiting() : ArrayList[Long]
  {
    return this.wawList;
  }
}
