package de.uks.plm.ftglb.actor.producerConsumer;

import x10.compiler.Inline;

/**
 * An implementation of a queue.
 */
public final class FixedRailQueue[T]{T haszero}
{
    private val internalStorage:Rail[T]{self!=null};
    private var N:Int;
    private var head:Int;
    private var tail:Int;

  /** Construct a fixed size queue */
    public def this(N:Int) : FixedRailQueue[T]
    { 
        this.N = N;
        this.internalStorage = new Rail[T](N);
        this.head = 0n;
        this.tail = 0n;
    }

    public def this(val that:FixedRailQueue[T]) : FixedRailQueue[T]
    {
        this.N = that.N;
        this.internalStorage = new Rail[T](that.internalStorage);
        this.head = that.head;
        this.tail = that.tail;
    }
    /** Check if the queue is empty */
    @Inline public def isEmpty() : Boolean
    {
        return (this.head == this.tail);
    }

    /** Add the element to the front of the queue. */
    @Inline public def push(val v:T) : void
    {
        // Add the element and increase the size
        this.internalStorage(this.tail++) = v;
    }

    /** Remove and return one element of the queue if FIFO order. */
    @Inline public def pop() : T
    {
        // Remove the first element from the queue.
        return this.internalStorage(this.head++);
    }

    /** Remove and return one element of the queue in LIFO order. */
    @Inline public def top() : T
    {
        return this.internalStorage(--this.tail);
    }

    /** Rewind. */
    @Inline public def rewind() : void
    {
        this.head = 0n;
    }

    /** Output the contents of the queue in the order they are stored */
    public def print () : void
    {
        Console.OUT.print("h = " + head + ", t = " + tail + ", ");
        Console.OUT.print ("[");
        for (var i:Int = this.head; i < this.tail; ++i)
        {
            Console.OUT.print (  ((i==this.head)?"":",")
                               + this.internalStorage(i));
        }
        Console.OUT.println ("]");
    }

    public static def main (val args:Rail[String]) : void
    {
        val myQueue = new FixedRailQueue[Int] (4n);
        myQueue.push (1n);
        myQueue.push (2n);
        myQueue.push (3n);
        myQueue.push (4n);
        myQueue.print ();

        myQueue.pop ();
        myQueue.pop ();
        myQueue.print ();

        myQueue.pop ();
        myQueue.pop ();
        myQueue.print ();

        myQueue.push (5n);
        myQueue.print ();

        myQueue.push (2n);
        myQueue.push (3n);
        myQueue.push (4n);
        myQueue.print ();
    }
}
