package de.uks.plm.ftglb.actor.producerConsumer;

import x10.util.Option;
import x10.util.OptionsParser;

import de.uks.plm.ftglb.actor.GLBParameters;
import de.uks.plm.ftglb.actor.GLB;

public final class ProducerConsumer
{
    public static def compute(val args:Rail[String]) : Rail[Long]
    {
        val opts:OptionsParser
            = new OptionsParser (args
                                , new Rail[Option]()
                                , [Option("d", "", "Depth")
                                   , Option("c", "", "consumers per producer")
                                   , Option("n"
                                            , ""
                                            ,  "Number of nodes to process "
                                             + "before probing. Default 200.")
                                   , Option("w"
                                            , ""
                                            ,  "Number of thieves to send out. "
                                             + " Default 1.")
                                   , Option("l", "", "Base of the lifeline")
                                   , Option("m", "", "Max potential victims")
                                   , Option("v", "", "Verbose. Default 0 (no).")
                                   , Option("k"
                                            , ""
                                            ,  "Number of cycles, before a "
                                             + "cyclic backup is written.")
                                   , Option("g"
                                             , ""
                                             ,  "Number of cycles, before a "
                                              + "monitor-call is started (all "
                                              + "r processRecorde")]);
        val d:Long = opts("-d", 64);
        val c:Long = opts("-c", 64);
        val n:Int = opts("-n", 512n);
        val l:Int = opts("-l", 32n);
        val m:Int = opts("-m", 1024n);
        val verbose:Int = opts("-v", GLBParameters.SHOW_RESULT_FLAG);
        val k:Int = opts("-k", 2048n);
        val g:Int = opts("-g", 128n);

        val P:Long = Place.numPlaces();

        var z0:Int = 1n;
        var zz:Int = l;
        while (zz < P)
        {
            z0++;
            zz *= l;
        }
        val z:Int = z0;
        val w:Int = opts("-w", z);

        Console.OUT.println(  "places = " + P
                            + "   d = " + d
                            + "   c = " + c
                            + "   w = " + w
                            + "   n = " + n
                            + "   l = " + l
                            + "   m = " + m
                            + "   z = " + z
                            + "   k = " + k
                            + "   g = " + g
                            + "   v = " + verbose);

        val init:() => Queue = () =>
                               {
                                   return (new Queue(c, d));
                               };
        val glb:GLB[Queue, Long] = new GLB[Queue, Long](init
                                                        , GLBParameters
                                                            (n
                                                            , w
                                                            , l
                                                            , z
                                                            , m
                                                            , k
                                                            , g
                                                            , verbose)
                                                        , true );
        val start:() => void = () =>
                               {
                                   glb.taskQueue().upper(0) = c + 1;
                                   glb.taskQueue().lower(0) = 0;
                                   glb.taskQueue().depth(0) = 0;
                                   ++glb.taskQueue().size;
                               };

        val result:Rail[Long] = glb.run(start);
        return (result);
    }

    public static def main(val args:Rail[String]) : void
    {
        compute(args);
    }
}
