package de.uks.plm.ftglb.actor.producerConsumer;

import x10.compiler.Inline;

import de.uks.plm.ftglb.actor.TaskQueue;
import de.uks.plm.ftglb.actor.TaskBag;

public class Queue implements TaskQueue[Queue, Long] {
    public val consumerRate:Long;
    public val maxDepth:Long;
    public var upper:Rail[Long];
    public var lower:Rail[Long];
    public var depth:Rail[Long];
    public var size:Long;
    private var count:Long;

    public def this(val consumerRate:Long, val maxDepth:Long) : Queue
    {
        this.consumerRate = consumerRate;
        this.maxDepth = maxDepth;
        this.upper = new Rail[Long](128);
        this.lower = new Rail[Long](128);
        this.depth = new Rail[Long](128);
        this.size = 0;
        this.count = 0;
    }

    public def this(val that:Queue)
    {
        this.consumerRate = that.consumerRate;
        this.maxDepth = that.maxDepth;
        this.upper = new Rail[Long](that.upper);
        this.lower = new Rail[Long](that.lower);
        this.depth = new Rail[Long](that.depth);
        this.size = that.size;
        this.count = that.count;
    }

    public def clone() : Queue
    {
        return (new Queue(this));
    }

    @Inline public def count() : Long
    {
        return (this.count);
    }

    @Inline public def getResult() : Result
    {
        return (new Result(this.count));
    }

    @Inline public def grow() : void
    {
        val newSize:Long = this.size * 2;
        val newUpper = new Rail[Long](newSize);
        val newLower = new Rail[Long](newSize);
        val newDepth = new Rail[Long](newSize);
        Rail.copy(this.upper, 0, newUpper, 0, this.size);
        Rail.copy(this.lower, 0, newLower, 0, this.size);
        Rail.copy(this.depth, 0, newDepth, 0, this.size);
        this.upper = newUpper;
        this.lower = newLower;
        this.depth = newDepth;
        this.size = newSize;
    }

    @Inline public def merge(val bag:TaskBag) : void
    {
        this.merge(bag as Bag);
    }

    @Inline public def merge(val bag:Bag) : void
    {
        val thisSize:Long = this.size();
        val bagSize:Long = bag.size();
        val newSize:Long = thisSize + bagSize;
        while (newSize > this.lower.size)
        {
            this.grow();
        }
        Rail.copy(bag.upper, 0, this.upper, thisSize, bagSize);
        Rail.copy(bag.lower, 0, this.lower, thisSize, bagSize);
        Rail.copy(bag.depth, 0, this.depth, thisSize, bagSize);

        this.size += bag.size();
    }

    @Inline public def merge(val that:TaskQueue[Queue, Long]) : void
    {
        this.merge(that as Queue);
    }

    @Inline public def merge(val that:Queue) : void
    {
        val thisSize:Long = this.size();
        val thatSize:Long = that.size();
        val newSize:Long = thisSize + thatSize;
        while (newSize > this.size)
        {
            this.grow();
        }

        Rail.copy(that.upper, 0, this.upper, thisSize, thatSize);
        Rail.copy(that.lower, 0, this.lower, thisSize, thatSize);
        Rail.copy(that.depth, 0, this.depth, thisSize, thatSize);

        this.size += that.size;
    }

    @Inline public def mergeResult(val that:TaskQueue[Queue, Long]) : void
    {
        this.mergeResult(that as Queue);
    }

    @Inline public def mergeResult(val that:Queue) : void
    {
        this.count += that.count;
    }

    @Inline public def printLog() : void {}

    @Inline public def process(val n:Long) : Boolean
    {
        var i:Long = 0;
        while(i++ < n)
        {
            if (0 == this.size)
            {
                return (false);
            }
            val top:Long = this.size - 1;
            val task:Long = this.upper(top) - 1;
            var depth:Long = this.depth(top);
            if (0 == task) // Producer
            {
                if (this.maxDepth > depth)
                {
                    ++depth;
                    this.depth(top) = depth;
                    this.lower(top) = 0;
                    val add:Long = this.consumerRate + 1;
                    this.upper(top) = add;
                    this.count += add;
                }
                else // if(this.maxDepth >= depth)
                {
                    --this.size;
                }
            }
            else // if (0 != task)
            {
                --this.upper(top);
                var sum:Long = 1;
                // Burn some time
                for (l:Long in 0..100)
                {
                    ++sum;
                }
                sum /= sum;
                this.count *= sum;
                if (this.lower(top) == this.upper(top))
                {
                    --this.size;
                }
            }
        }
        return (this.size != 0);
    }

    @Inline public def setResult(val that:TaskQueue[Queue, Long]) : void
    {
        this.setResult(that as Queue);
    }

    @Inline public def setResult(val that:Queue) : void
    {
        this.count = that.count;
    }

    @Inline public def size() : Long
    {
        return (this.size);
    }

    @Inline public def split() : Bag
    {
        var s:Long = 0;
        for (var i:Long=0; i < size; ++i)
        {
            if ((upper(i) - lower(i)) >= 2)
            {
                ++s;
            }
        }

        if (0 == s) 
        {
            return (null);
        }

        val bag = new Bag(s);
        s = 0;
        for (var i:Long = 0; i < size; ++i)
        {
            val p = this.upper(i) - this.lower(i);
            if (2n <= p)
            {
                bag.depth(s) = this.depth(i);
                bag.lower(s) = this.lower(i);
                this.lower(i) += (p / 2n);
                bag.upper(s++) = this.lower(i);
            }
        }
        return (bag);
    }
}
