package de.uks.plm.utils;

import x10.compiler.Inline;

public class RingBuffer[T] {T haszero}
{
    public static DEFAULT_SIZE:Long = 10;
    /**
     * Buffer to store elements
     */
    // mb: from protected to public
    public var buffer:Rail[T];

    /**
     * Start of the buffer
     * Always between 0 and buffer.size - 1.
     */
    private var start:Long;

    /**
     * End of the buffer (next free index).
     * Always between 0 and buffer.size - 1.
     */
    private var end:Long;

    /**
     * Number of elements in the buffer.
     * Always between 0 and buffer.size - 1.
     */
    private var size:Long;

    public def this() : RingBuffer[T]
    {
        this(RingBuffer.DEFAULT_SIZE);
    }

    public def this(val size:Long) : RingBuffer[T]
    {
        this.buffer = new Rail[T](size);
        this.start = 0;
        this.end = 0;
        this.size = 0;
    }

    public def this(val buffer:Rail[T]) : RingBuffer[T]
    {
        this(buffer, 0);
    }

    public def this(val buffer:Rail[T], val start:Long) : RingBuffer[T]
    {
        this(buffer, start, start);
        this.size = buffer.size;
    }

    public def this(val buffer:Rail[T], val start:Long, val end:Long)
        : RingBuffer[T]
    {
        if (0 == buffer.size)
        {
            this.buffer = new Rail[T](0);
            this.start = 0;
            this.end = 0;
            this.size = 0;
        }
        else
        {
            assert (0 <= start && buffer.size > start)
                : here + ": start must be in [0, " + (buffer.size - 1) + "].";

            assert (0 <= end && buffer.size > end)
                : here + ": end must be in [0, " + (buffer.size - 1) + "].";

            this.buffer = buffer;
            this.start = start;
            this.end = end;
            this.size = (buffer.size + end - start) % buffer.size;
        }
    }

    public def this(val that:RingBuffer[T]) : RingBuffer[T]
    {
        this.buffer = new Rail[T](that.buffer);
        this.start = that.start;
        this.end = that.end;
        this.size = that.size;
    }

    // Commented because: http://jira.codehaus.org/browse/XTENLANG-3317
    // @Inline public def clone() : RingBuffer[T]
    // {
    //     return new RingBuffer[T](this);
    // }

    @Inline public def end():Long
    {
        return (this.end);
    }

    // Gibt bis die ersten n Element vom start aus geszählt zurück.
    @Inline public def peekFromBottom(val n:Long) : Rail[T]
    {
        val toCopy:Long = Math.min(n, this.size());
        val result:Rail[T] = new Rail[T](toCopy);
        val firstCopy = Math.min(toCopy, this.buffer.size - this.start);
        Rail.copy(this.buffer, this.start, result, 0, firstCopy);
        if (firstCopy < toCopy)
        {
            val secondCopy:Long = toCopy - firstCopy;
            Rail.copy(this.buffer, 0, result, firstCopy, secondCopy);
        }
        
        return (result);
    }

    @Inline public def getFromBottom(val n:Long) : Rail[T]
    {
        val toCopy:Long = Math.min(n, this.size());
        if (0 == toCopy)
        {
            return (new Rail[T](toCopy));
        }
        val result:Rail[T] = this.peekFromBottom(toCopy);
        if (toCopy == this.size)
        {
            this.size = 0;
            this.start = 0;
            this.end = 0;
            return (result);
        }
        else
        {
            this.start += toCopy;
            this.start %= this.buffer.size;
            this.size -= toCopy;
        }
        return (result);
    }

    @Inline public def peekFromTop(val n:Long) : Rail[T]
    {
        val toCopy:Long = Math.min(n, this.buffer.size);
        val result:Rail[T] = new Rail[T](toCopy);
        val firstCopy = Math.min(toCopy, this.end);
        val secondCopy = toCopy - firstCopy;
        Rail.copy(this.buffer
                  , this.end - firstCopy
                  , result
                  , secondCopy
                  , firstCopy);
        if (0 < secondCopy)
        {
            Rail.copy(this.buffer
                      , this.buffer.size - secondCopy
                      , result
                      , 0
                      , secondCopy);
        }
        return (result);
    }

    @Inline public def getFromTop(val n:Long) : Rail[T]
    {
        val toCopy:Long = Math.min(n, this.buffer.size);
        if (0 == toCopy)
        {
            return (new Rail[T](toCopy));
        }
        val result:Rail[T] = this.peekFromTop(toCopy);
        if (toCopy == this.size)
        {
            this.start = 0;
            this.end = 0;
            this.size = 0;
        }
        else
        {
            this.end -= toCopy;
            this.end += this.buffer.size;
            this.end %= this.buffer.size;
            this.size -= toCopy;
        }
        return (result);
    }

    @Inline public def grow() : void
    {
        val newSize:Long = Math.max(this.buffer.size * 2, 1);
        val newBuffer:Rail[T] = new Rail[T](newSize);
        val toCopy = this.size;
        val firstCopy = Math.min(this.buffer.size - this.start, this.size);
        Rail.copy(this.buffer, start, newBuffer, 0, firstCopy);
        if (firstCopy < toCopy)
        {
            val secondCopy:Long = toCopy - firstCopy;
            Rail.copy(this.buffer, 0, newBuffer, firstCopy, secondCopy);
        }
        this.buffer = newBuffer;
        this.start = 0;
        this.end = Math.max(this.size, 0);
    }

    @Inline public def normalize() : void
    {
        if (0 != this.start || this.buffer.size != this.size)
        {
            val newBuffer:Rail[T] = new Rail[T](this.size);
            val toCopy:Long = this.size;
            val firstCopy:Long = Math.min(this.buffer.size - this.start, toCopy);
            Rail.copy(this.buffer, this.start, newBuffer, 0, firstCopy);
            if (firstCopy < toCopy)
            {
                val secondCopy:Long = toCopy - firstCopy;
                Rail.copy(this.buffer, 0, newBuffer, firstCopy, secondCopy);
            }
            this.buffer = newBuffer;
            this.start = 0;
            this.end = this.size % this.buffer.size;
            assert (   0 == this.start
                    && 0 == this.end
                    && this.size == this.buffer.size)
                    : here + ": normalization failed.";
        }
    }

//        @Inline public def mergeAtTop(val that:RingBuffer[T]) : void
//        {
//            if (0 == that.size())
//            {
//                return;
//            }
//            /* else */ if (0 == this.size())
//            {
//                this.buffer = that.buffer;
//                this.start = that.start;
//                this.end = that.end;
//                this.size = that.size;
//                return;
//            }
//            // else
//            // {
//                val thisSize:Long = this.size();
//                val thatSize:Long = that.size();
//                val newSize:Long = thisSize + thatSize;
// 
//                this.normalize();
//                that.normalize();
// 
//                while (newSize > this.buffer.size)
//                {
//                    this.grow();
//                }
//                Rail.copy(that.buffer, 0, this.buffer, this.end, that.end);
//                this.end = that.end;
//                this.size = newSize;
//            // }
//        }

    @Inline public def mergeAtTop(val that:RingBuffer[T]) : void
    {
        if (0 == that.size())
        {
            return;
        }
        if (0 == this.size())
        {
            this.buffer = new Rail[T](that.buffer);
            this.start = that.start;
            this.end = that.end;
            this.size = that.size;
            return;
        }
        val thisSize:Long = this.size;
        val thatSize:Long = that.size;
        val newSize:Long = thisSize + thatSize;
        if (newSize > this.buffer.size)
        {
            val newBuffer:Rail[T] = new Rail[T](newSize);

            val firstThisCopy:Long
                = Math.min(this.buffer.size - this.start, thisSize);

            Rail.copy(this.buffer, this.start, newBuffer, 0, firstThisCopy);
            if (firstThisCopy < thisSize)
            {
                val secondThisCopy:Long = thisSize - firstThisCopy;
                Rail.copy(this.buffer
                          , 0
                          , newBuffer
                          , firstThisCopy
                          , secondThisCopy);
            }

            val firstThatCopy:Long
                = Math.min(that.buffer.size - that.start, thatSize);
            Rail.copy(that.buffer
                      , that.start
                      , newBuffer
                      , thisSize
                      , firstThatCopy);
            if (firstThatCopy < thatSize)
            {
                val secondThatCopy:Long = thatSize - firstThatCopy;
                Rail.copy(that.buffer
                          , 0
                          , newBuffer
                          , thisSize + firstThatCopy
                          , secondThatCopy);
            }
            this.buffer = newBuffer;
            this.start = 0;
            this.end = 0;
            this.size = newSize;
        }
        else // if (newSize <= this.buffer.size)
        {
            var copied:Long = 0;
            do
            {
                val thatStartIdx:Long
                    = (that.start + copied) % that.buffer.size;

                val thatRest:Long = thatSize - copied;
                var copy:Long
                    = Math.min(thatRest, that.buffer.size - thatStartIdx);

                copy = Math.min(copy, this.buffer.size - this.end);

                Rail.copy(that.buffer
                          , thatStartIdx
                          , this.buffer
                          , this.end
                          , copy);

                copied += copy;
                this.end += copy;
                this.end %= this.buffer.size;
                this.size += copy;
                assert (0 <= thatRest)
                       : here + ": buffer-size got negative. "
                         + "Something terrible happened.";
            }
            while(copied < thatSize);
            assert (thatSize == copied)
                   : here + ": Wrong amount copied. "
                     + "Something terrible happened.";
        }
    }

    @Inline public def mergeAtBottom(val that:RingBuffer[T]) : void
    {
        if (0 == that.size())
        {
            return;
        }
        else
        {
            val thatClone = new RingBuffer[T](that);
            if (0 == this.size())
            {
                this.buffer = new Rail[T](that.buffer);
            }
            else
            {
                thatClone.mergeAtTop(this);
                this.buffer = new Rail[T](thatClone.buffer);
            }
            this.start = thatClone.start;
            this.end = thatClone.end;
            this.size = thatClone.size;
        }
    }

    @Inline public def peekTop() : T
    {
        assert (0 < this.size) : here + ": cannot pop from an empty Stack!";
        return this.buffer((end - 1 + this.buffer.size) % this.buffer.size);
        
    }

    @Inline public def popTop() : T
    {
        val value = this.peekTop();
        --this.size;
        if (0 >= this.size)
        {
            this.start = 0;
            this.end = 0;
        }
        else // if (0 < this.size)
        {
            --this.end;
            this.end += this.buffer.size;
            this.end %= this.buffer.size;
        }
        return (value);
    }

    @Inline public def peekBottom() : T
    {
        return (this.buffer(this.start));
    }

    @Inline public def popBottom():T
    {
        assert (0 < this.size) : here + ": cannot pop from an empty Stack!";
        --this.size;
        val value:T = this.buffer(start);
        ++this.start;
        this.start %= this.buffer.size;
        if (0 == this.size)
        {
            this.start = 0;
            this.end = 0;
        }
        return (value);
    }

    @Inline public def pushBottom(val value:T) : void
    {
        if (this.buffer.size <= this.size)
        {
            this.grow();
        }
        this.start += (this.buffer.size - 1);
        this.start %= this.buffer.size;
        this.buffer(start) = value;
        ++this.size;
    }

    @Inline public def pushBottom(val values:Rail[T]) : void
    {
        val toCopy:Long = values.size;
        val newCount:Long = this.size + toCopy;
        while (newCount > this.buffer.size)
        {
            this.grow();
        }
        val firstCopy:Long = Math.min(this.start, toCopy);

        Rail.copy(values
                  , toCopy - firstCopy
                  , this.buffer
                  , start-firstCopy
                  , firstCopy);

        if (firstCopy < toCopy)
        {
            val secondCopy:Long = toCopy - firstCopy;
            Rail.copy(values
                      , 0
                        , this.buffer
                        , this.buffer.size - secondCopy
                        , secondCopy);
        }
        this.size += toCopy;
        this.start -= toCopy;
        this.start += this.buffer.size;
        this.start %= this.buffer.size;
    }

    @Inline public def pushTop(val value:T) : void
    {
        if (this.buffer.size <= this.size)
        {
            this.grow();
        }
        this.buffer(this.end++) = value;
        this.end %= this.buffer.size;
        ++this.size;
    }

    @Inline public def pushTop(val value:Rail[T]) : void
    {
        val toCopy:Long = value.size;
        val newCount:Long = this.size + toCopy;
        while (newCount > this.buffer.size)
        {
            this.grow();
        }
        val firstCopy:Long = Math.min(this.buffer.size - this.end, toCopy);
        Rail.copy(value, 0, this.buffer, end, firstCopy);
        if (firstCopy < toCopy)
        {
            val secondCopy:Long = toCopy - firstCopy;
            Rail.copy(value, firstCopy, this.buffer, 0, secondCopy);
        }
        this.end += toCopy;
        if (0 < this.buffer.size)
        {
            this.end %= this.buffer.size;
        }
        this.size += toCopy;
    }

    @Inline public def size() : Long
    {
        return (this.size);
    }

    @Inline public def start() : Long
    {
        return (this.start);
    }

    @Inline public def toString() : String
    {
        var output:String = "[";
        for (var idx:Long = 0; idx < this.buffer.size; ++idx)
        {
            if (this.start == idx)
            {
                output += "B";
            }

            output += this.buffer(idx);

            if (((this.buffer.size + this.end) % this.buffer.size) == idx)
            {
                output += "T";
            }
            if (idx != this.buffer.size -1)
            {
                output += ", ";
            }
        }
        output += "]";
        output += "\nSize: "+ this.size;
        return (output);
    }

    public static def main(val args:Rail[String]) : void
    {
        val d1:Rail[Any] = new Rail[Any](16, null as Any);
        d1(14) = 0;
        d1(15) = 1;
        d1(0) = 2;
        d1(1) = 3;
        d1(2) = 4;
        d1(3) = 5;
        d1(4) = 6;

        val d2:Rail[Any] = new Rail[Any](4);
        d2(0) = 100;
        d2(1) = 101;
        d2(2) = 102;
        d2(3) = 103;

        var buffer1:RingBuffer[Any] = new RingBuffer[Any](d1, 14, 5);
        var buffer2:RingBuffer[Any] = new RingBuffer[Any](d2);

        Console.OUT.println(buffer1);
        Console.OUT.println();
        Console.OUT.println(buffer2);
        Console.OUT.println();

        var buffer3:RingBuffer[Any] = new RingBuffer[Any](buffer1);
        buffer3.mergeAtBottom(buffer2);

        var buffer4:RingBuffer[Any] = new RingBuffer[Any](buffer2);
        buffer4.mergeAtTop(buffer1);

        Console.OUT.println(buffer3);
        Console.OUT.println();
        Console.OUT.println(buffer4);
    }
}
