package de.uks.plm.utils;

import x10.compiler.Inline;

import x10.util.ArrayList;
import x10.util.Collection;

public class SimpleStack[T] {T haszero}
{
    public static val DEFAULT_SIZE = 1024;

    private var data:Rail[T];
    private var size:Long;

    public static def main(val args:Rail[String]) : void
    {
        val st:SimpleStack[Long] = new SimpleStack[Long]();
        for (value:Long in 0..10)
        {
            st.push(value);
        }
        Console.OUT.println(st);
        Console.OUT.println(st.peek());
        Console.OUT.println(st.peek(5));
        Console.OUT.println(st.pop(12));
        while (!st.isEmpty())
        {
            Console.OUT.println(st.pop());
            Console.OUT.println(st);
        }
        try
        {
            st.pop();
        }
        catch (val e:AssertionError)
        {
            Console.OUT.println(e);
        }
    }

    @Inline public def this() : SimpleStack[T]
    {
        this(SimpleStack.DEFAULT_SIZE);
    }

    @Inline public def this(val size:Long) : SimpleStack[T]
    {
        this.data = new Rail[T](size);
        this.size = 0;
    }

    @Inline public def this(val data:Rail[T]) : SimpleStack[T]
    {
        this.data = new Rail[T](data.size);
        Rail.copy(data, 0, this.data, 0, data.size);
        this.size = data.size;
    }

    @Inline public def this(val stack:SimpleStack[T]) : SimpleStack[T]
    {
        this(stack.data);
    }

    @Inline public def clone() : SimpleStack[T]
    {
        return (new SimpleStack[T](this));
    }

    @Inline private def grow() : void
    {
        assert (this.data.size == this.size) : here + ": stack not full!";
        val data:Rail[T] = new Rail[T](2 * this.size);
        Rail.copy(this.data, 0, data, 0, this.size);
        this.data = data;
    }

    @Inline public def peek() : T
    {
        assert (!this.isEmpty()) : here + ": cannot peek on an empty stack!";

        return (this.data(this.size - 1));
    }

    @Inline public def isEmpty() : Boolean
    {
        return (0 == this.size);
    }

    @Inline public def peek(val n:Long) : Rail[T]
    {
        assert (!this.isEmpty()) : here + ": cannot peek on an empty stack!";

        assert (n - 1 <= this.size)
               : here + ": n must be in [0, " + this.size + "]!";

        val res:Rail[T] = new Rail[T](n);
        Rail.copy(this.data, this.size - n, res, 0, n);
        return (res);
    }

    @Inline public def pop() : T
    {
        assert (!this.isEmpty())
               : here + ": cannot pop something from an empty stack!";

        return (this.data(--this.size));
    }

    @Inline public def pop(val n:Long) : Rail[T]
    {
        assert (!this.isEmpty())
               : here + ": cannot pop something from an empty stack!";

        assert (n  <= this.size)
               : here + ": n must be in [0, " + this.size + "]!";

        val result:Rail[T] = new Rail[T](n);
        Rail.copy(this.data, this.size - n, result, 0, n);
        this.size -= n;
        return (result);
    }

    @Inline public def push(val item:T) : void
    {
        if (this.data.size == this.size)
        {
            this.grow();
        }
        this.data(this.size++) = item;
    }

    @Inline public def pushAll(val items:Rail[T]) : void
    {
        while((this.size + items.size) > this.data.size)
        {
            this.grow();
        }
        Rail.copy(items, 0, this.data, this.size, items.size);
    }

    @Inline public def pushAll(val items:ArrayList[T]) : void
    {
        this.pushAll(items.toRail());
    }

    @Inline public def pushAll(val that:SimpleStack[T]) : void
    {
        while (this.size + that.size > this.data.size)
        {
            this.grow();
        }
        Rail.copy(that.data, 0, this.data, this.size, that.size);
        this.size += that.size;
    }

    @Inline public def pushAll(val items:Collection[T]) : void
    {
        while((this.size + items.size()) > this.data.size)
        {
            this.grow();
        }

        for(item:T in items)
        {
            this.data(this.size++) = item;
        }
    }

    @Inline public def size() : Long
    {
        return (this.size);
    }

    @Inline public def toString() : String
    {
        var res:String = "[";
        for (index:Long in 0..(this.size - 1))
        {
            res += this.data(index);
            if (index < this.size - 1)
            {
                res+= ", ";
            }
        }
        res += "]";
        return (res);
    }
}
